# Copyright (c) 2010-2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock
import os
import tempfile
import unittest

import io

from swiftbench import utils


class TestUtils(unittest.TestCase):

    @mock.patch.object(utils, "TRUE_VALUES")
    def test_config_true_value(self, mocked):
        utils.TRUE_VALUES = 'hello world'.split()
        for val in 'hello world HELLO WORLD'.split():
            self.assertTrue(utils.config_true_value(val) is True)
            self.assertTrue(utils.config_true_value(True) is True)
            self.assertTrue(utils.config_true_value('foo') is False)
            self.assertTrue(utils.config_true_value(False) is False)

    def test_readconf(self):
        conf = '''[section1]
foo = bar

[section2]
log_name = yarr'''
        fd, temppath = tempfile.mkstemp(dir='/tmp')
        with os.fdopen(fd, 'w') as f:
            f.write(conf)
        # Test both a real file and a file-like object
        for conffile in (temppath, io.StringIO(conf)):
            result = utils.readconf(conffile)
            expected = {'__file__': conffile,
                        'log_name': None,
                        'section1': {'foo': 'bar'},
                        'section2': {'log_name': 'yarr'}}
            self.assertEqual(result, expected)

            result = utils.readconf(conffile, 'section1')
            expected = {'__file__': conffile, 'log_name': 'section1',
                        'foo': 'bar'}
            self.assertEqual(result, expected)

            result = utils.readconf(conffile,
                                    'section2').get('log_name')
            expected = 'yarr'
            self.assertEqual(result, expected)

            result = utils.readconf(conffile, 'section1',
                                    log_name='foo').get('log_name')
            expected = 'foo'
            self.assertEqual(result, expected)

            result = utils.readconf(conffile, 'section1',
                                    defaults={'bar': 'baz'})
            expected = {'__file__': conffile, 'log_name': 'section1',
                        'foo': 'bar', 'bar': 'baz'}
            self.assertEqual(result, expected)
        self.assertRaises(SystemExit, utils.readconf, temppath, 'section3')
        os.unlink(temppath)
        self.assertRaises(SystemExit, utils.readconf, temppath)

    def test_readconf_raw(self):
        conf = '''[section1]
foo = bar

[section2]
log_name = %(yarr)s'''
        fd, temppath = tempfile.mkstemp(dir='/tmp')
        with os.fdopen(fd, 'w') as f:
            f.write(conf)
        # Test both a real file and a file-like object
        for conffile in (temppath, io.StringIO(conf)):
            result = utils.readconf(conffile, raw=True)
            expected = {'__file__': conffile,
                        'log_name': None,
                        'section1': {'foo': 'bar'},
                        'section2': {'log_name': '%(yarr)s'}}
            self.assertEqual(result, expected)
        os.unlink(temppath)
        self.assertRaises(SystemExit, utils.readconf, temppath)

    @mock.patch("swiftbench.utils.getproxies")
    @mock.patch("swiftbench.utils.proxy_bypass")
    def test_using_http_proxy(self, mock_proxy_bypass, mock_getproxies):
        mock_getproxies.return_value = {'http': 'proxy', 'https': 'proxy'}

        def fake_proxy_bypass(url):
            return url == "localhost"
        mock_proxy_bypass.side_effect = fake_proxy_bypass

        self.assertTrue(utils.using_http_proxy("http://host1/"))
        self.assertFalse(utils.using_http_proxy("http://localhost/"))
        self.assertTrue(utils.using_http_proxy("https://host1/"))
        self.assertFalse(utils.using_http_proxy("https://localhost/"))
        self.assertFalse(utils.using_http_proxy("dummy://localhost/"))
        self.assertFalse(utils.using_http_proxy("dummy://host1/"))

    def test_get_size_bytes(self):
        self.assertEqual(utils.get_size_bytes('10M'), 10 * 1024 * 1024)
        self.assertEqual(utils.get_size_bytes('4k'), 4096)
        self.assertEqual(utils.get_size_bytes('5G'), 5 * 1024 * 1024 * 1024)
        self.assertEqual(utils.get_size_bytes('1234'), 1234)
        self.assertEqual(utils.get_size_bytes('1 k'), 1024)
        self.assertEqual(utils.get_size_bytes(' 1k'), 1024)
        self.assertEqual(utils.get_size_bytes('1k '), 1024)

        with self.assertRaises(ValueError):
            utils.get_size_bytes('1K')
        with self.assertRaises(ValueError):
            utils.get_size_bytes('1m')
        with self.assertRaises(ValueError):
            utils.get_size_bytes('1g')

        with self.assertRaises(ValueError):
            utils.get_size_bytes('1kb')
        with self.assertRaises(ValueError):
            utils.get_size_bytes('1kB')

        with self.assertRaises(ValueError):
            utils.get_size_bytes('1T')
        with self.assertRaises(ValueError):
            utils.get_size_bytes('1P')
        with self.assertRaises(ValueError):
            utils.get_size_bytes('1E')
        with self.assertRaises(ValueError):
            utils.get_size_bytes('1Y')
        with self.assertRaises(ValueError):
            utils.get_size_bytes('asdf')

        self.assertEqual(utils.get_size_bytes(1), 1)
        with self.assertRaises(TypeError):
            utils.get_size_bytes(1.0)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import io
import json
import re
import sys
import uuid
import time
import random
import signal
import socket
import logging
from contextlib import contextmanager, closing
from optparse import Values

import eventlet
import eventlet.pools
from eventlet.green.httplib import CannotSendRequest

import requests.exceptions

import swiftclient as client

from swiftbench.utils import config_true_value, using_http_proxy, \
    get_size_bytes


try:
    from swift.common import direct_client
except ImportError:
    direct_client = None

HTTP_CONFLICT = 409


def _func_on_containers(logger, conf, concurrency_key, func, **kwargs):
    """Run a function on each container with concurrency."""

    bench = Bench(logger, conf, [])
    pool = eventlet.GreenPool(int(getattr(conf, concurrency_key)))
    for container in conf.containers:
        pool.spawn_n(func, bench.url, bench.token, container, **kwargs)
    pool.waitall()


def delete_containers(logger, conf):
    """Utility function to delete benchmark containers."""

    def _deleter(url, token, container):
        try:
            client.delete_container(url, token, container)
        except (client.ClientException,
                requests.exceptions.ConnectionError) as e:
            if e.http_status != HTTP_CONFLICT:
                logger.warn("Unable to delete container '%s'. "
                            "Got http status '%d'."
                            % (container, e.http_status))

    _func_on_containers(logger, conf, 'del_concurrency', _deleter)


def create_containers(logger, conf):
    """Utility function to create benchmark containers."""

    if conf.policy_name:
        logger.info("Creating containers with storage policy: %s" %
                    conf.policy_name)
        _func_on_containers(logger, conf, 'put_concurrency',
                            client.put_container,
                            headers={'X-Storage-Policy': conf.policy_name})
    else:
        _func_on_containers(logger, conf, 'put_concurrency',
                            client.put_container)


class SourceFile(object):
    """
    Iterable, file-like object to lazily emit a bunch of zeros in
    reasonable-size chunks.

    swift.common.direct_client wants iterables, but swiftclient wants
    file-like objects where hasattr(thing, 'read') is true. Therefore,
    this class can do both.
    """

    def __init__(self, size, chunk_size=1024 * 64):
        self.pos = 0
        self.size = size
        self.chunk_size = chunk_size

    def __iter__(self):
        return self

    def __len__(self):
        return self.size

    def next(self):
        if self.pos >= self.size:
            raise StopIteration
        chunk_size = min(self.size - self.pos, self.chunk_size)
        self.pos += chunk_size
        return b'0' * chunk_size

    def read(self, desired_size):
        chunk_size = min(self.size - self.pos, desired_size)
        self.pos += chunk_size
        return b'0' * chunk_size


class ConnectionPool(eventlet.pools.Pool):

    def __init__(self, url, size):
        self.url = url
        eventlet.pools.Pool.__init__(self, size, size)

    def create(self):
        return client.http_connection(self.url)


class BenchServer(object):
    """
    A BenchServer binds to an IP/port and listens for bench jobs.  A bench
    job consists of the normal conf "dict" encoded in JSON, terminated with an
    EOF.  The log level is at least INFO, but DEBUG may also be specified in
    the conf dict.

    The server will wait forever for jobs, running them one at a time.
    """
    def __init__(self, logger, bind_ip, bind_port):
        self.logger = logger
        self.bind_ip = bind_ip
        self.bind_port = int(bind_port)

    def run(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.logger.info('Binding to %s:%s', self.bind_ip, self.bind_port)
        s.bind((self.bind_ip, self.bind_port))
        s.listen(20)
        while True:
            client, address = s.accept()
            self.logger.debug('Accepting connection from %s:%s', *address)
            client_file = client.makefile('rwb', 1)
            client_file = io.TextIOWrapper(client_file)
            json_data = client_file.read()
            conf = Values(json.loads(json_data))

            self.logger.info(
                'Starting run for %s:%s [put/get/del_concurrency: %s/%s/%s, '
                'num_objects: %s, num_gets: %s]', address[0], address[1],
                conf.put_concurrency, conf.get_concurrency,
                conf.del_concurrency, conf.num_objects, conf.num_gets)

            logger = logging.getLogger('bench-server')
            level = logging.DEBUG if conf.log_level.lower() == 'debug' \
                else logging.INFO
            logger.setLevel(level)
            loghandler = logging.StreamHandler(client_file)
            logformat = logging.Formatter(
                'bench-server %(asctime)s %(levelname)s %(message)s')
            loghandler.setFormatter(logformat)
            logger.addHandler(loghandler)

            controller = BenchController(logger, conf)
            try:
                controller.run()
            except socket.error:
                logger.warning('Socket error', exc_info=1)

            logger.removeHandler(loghandler)
            client_file.close()
            client.close()

            self.logger.info('...bench run completed; waiting for next run.')


class Bench(object):

    def __init__(self, logger, conf, names):
        self.logger = logger
        self.aborted = False
        self.user = conf.user
        self.key = conf.key
        self.auth_url = conf.auth
        self.use_proxy = config_true_value(conf.use_proxy)
        if not self.use_proxy and direct_client is None:
            self.logger.critical("You need to have swift installed if you are "
                                 "not using the proxy")
            sys.exit(1)
        self.auth_version = conf.auth_version
        self.logger.info("Auth version: %s" % self.auth_version)
        if self.use_proxy:
            if using_http_proxy(self.auth_url):
                logger.warn("Auth is going through HTTP proxy server. This "
                            "could affect test result")
            url, token = client.get_auth(self.auth_url, self.user, self.key,
                                         auth_version=self.auth_version)
            self.token = token
            self.account = url.split('/')[-1]
            if conf.url == '':
                self.url = url
            else:
                self.url = conf.url
        else:
            self.token = 'SlapChop!'
            self.account = conf.account
            self.url = conf.url
            self.ip, self.port = self.url.split('/')[2].split(':')

        if using_http_proxy(self.url):
            logger.warn("Communication with Swift server is going through "
                        "HTTP proxy server. This could affect test result")

        self.object_size = get_size_bytes(conf.object_size)
        self.object_sources = conf.object_sources
        self.lower_object_size = get_size_bytes(conf.lower_object_size)
        self.upper_object_size = get_size_bytes(conf.upper_object_size)
        self.files = []
        if self.object_sources:
            self.object_sources = self.object_sources.split()
            self.files = []
            for f in self.object_sources:
                with open(f, 'rb') as fp:
                    self.files.append(fp.read())

        self.put_concurrency = int(conf.put_concurrency)
        self.get_concurrency = int(conf.get_concurrency)
        self.del_concurrency = int(conf.del_concurrency)
        self.total_objects = int(conf.num_objects)
        self.total_gets = int(conf.num_gets)
        self.timeout = int(conf.timeout)
        self.devices = conf.devices.split()
        self.names = names
        self.conn_pool = ConnectionPool(self.url,
                                        max(self.put_concurrency,
                                            self.get_concurrency,
                                            self.del_concurrency))

    def _log_status(self, title):
        total = time.time() - self.beginbeat
        self.logger.info(
            '%(complete)s %(title)s [%(fail)s failures], %(rate).01f/s',
            {'title': title, 'complete': self.complete,
             'fail': self.failures,
             'rate': (float(self.complete) / total)})

    @contextmanager
    def connection(self):
        try:
            hc = self.conn_pool.get()
            try:
                yield hc
            except CannotSendRequest:
                self.logger.info("CannotSendRequest.  Skipping...")
                try:
                    hc.close()
                except Exception:
                    pass
                self.failures += 1
                hc = self.conn_pool.create()
        finally:
            self.conn_pool.put(hc)

    def run(self):
        pool = eventlet.GreenPool(self.concurrency)
        self.beginbeat = self.heartbeat = time.time()
        self.heartbeat -= 13    # just to get the first report quicker
        self.failures = 0
        self.complete = 0
        for i in range(self.total):
            if self.aborted:
                break
            pool.spawn_n(self._run, i)
        pool.waitall()
        self._log_status(self.msg + ' **FINAL**')

    def _run(self, thread):
        return


class DistributedBenchController(object):
    """
    This class manages a distributed swift-bench run.  For this Controller
    class to make sense, the conf.bench_clients list must contain at least one
    entry.

    The idea is to split the configured load between one or more
    swift-bench-client processes, each of which use eventlet for concurrency.
    We deliberately take a simple, naive approach with these limitations:
        1) Concurrency, num_objects, and num_gets are spread evenly between the
           swift-bench-client processes.  With a low concurrency to
           swift-bench-client count ratio, rounding may result in a greater
           than desired aggregate concurrency.
        2) Each swift-bench-client process runs independently so some may
           finish up before others, i.e. the target aggregate concurrency is
           not necessarily present the whole time.  This may bias aggregate
           reported rates lower than a more efficient architecture.
        3) Because of #2, some swift-bench-client processes may be running GETs
           while others are still runinng their PUTs.  Because of this
           potential skew, distributed runs will not isolate one operation at a
           time like a single swift-bench run will.
        3) Reported aggregate rates are simply the sum of each
           swift-bench-client process reported FINAL number.  That's probably
           inaccurate somehow.
    """

    def __init__(self, logger, conf):
        self.logger = logger
        # ... INFO 1000 PUTS **FINAL** [0 failures], 34.9/s
        self.final_re = re.compile(
            r'INFO (\d+) (.*) \*\*FINAL\*\* \[(\d+) failures\], (\d+\.\d+)/s')
        self.clients = conf.bench_clients
        del conf.bench_clients
        for key, minval in [('put_concurrency', 1),
                            ('get_concurrency', 1),
                            ('del_concurrency', 1),
                            ('num_objects', 0),
                            ('num_gets', 0)]:
            setattr(conf, key,
                    max(minval, int(getattr(conf, key)) / len(self.clients)))
        self.conf = conf

    def run(self):
        eventlet.patcher.monkey_patch(socket=True)
        pool = eventlet.GreenPool(size=len(self.clients))
        pile = eventlet.GreenPile(pool)
        for c in self.clients:
            pile.spawn(self.do_run, c)
        results = {
            'PUTS': dict(count=0, failures=0, rate=0.0),
            'GETS': dict(count=0, failures=0, rate=0.0),
            'DEL': dict(count=0, failures=0, rate=0.0),
        }
        for result in pile:
            for k, v in result.items():
                target = results[k]
                target['count'] += int(v['count'])
                target['failures'] += int(v['failures'])
                target['rate'] += float(v['rate'])
        for k in ['PUTS', 'GETS', 'DEL']:
            v = results[k]
            self.logger.info('%d %s **FINAL** [%d failures], %.1f/s' % (
                v['count'], k, v['failures'], v['rate']))

    def do_run(self, client):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        ip, port = client.split(':')
        s.connect((ip, int(port)))
        s.sendall(json.dumps(self.conf.__dict__).encode('ascii'))
        s.shutdown(socket.SHUT_WR)
        s_file = s.makefile('rb', 1)
        result = {}
        for line in s_file:
            line = line.decode('ascii')
            match = self.final_re.search(line)
            if match:
                g = match.groups()
                result[g[1]] = {
                    'count': g[0],
                    'failures': g[2],
                    'rate': g[3],
                }
            else:
                sys.stderr.write('%s %s' % (client, line))
        return result


class BenchController(object):

    def __init__(self, logger, conf):
        self.logger = logger
        self.conf = conf
        self.names = []
        self.delete = config_true_value(conf.delete)
        self.gets = int(conf.num_gets)
        self.aborted = False
        self.delay = int(self.conf.delay)

    def sigint1(self, signum, frame):
        if self.delete:
            print(
                'SIGINT received; finishing up and running DELETE.\n'
                'Send one more SIGINT to exit *immediately*.',
                file=sys.stderr)
            self.aborted = True
            if self.running and not isinstance(self.running, BenchDELETE):
                self.running.aborted = True
            signal.signal(signal.SIGINT, self.sigint2)
        else:
            self.sigint2(signum, frame)

    def sigint2(self, signum, frame):
        sys.exit('Final SIGINT received.')

    def run(self):
        eventlet.patcher.monkey_patch(socket=True)
        signal.signal(signal.SIGINT, self.sigint1)
        puts = BenchPUT(self.logger, self.conf, self.names)
        self.running = puts
        puts.run()
        if self.gets and not self.aborted:
            gets = BenchGET(self.logger, self.conf, self.names)
            self.running = gets
            gets.run()
        if self.delete:
            if self.delay != 0:
                self.logger.info('Delay before '
                                 'DELETE request %s sec'
                                 % self.delay)
                time.sleep(self.delay)
            dels = BenchDELETE(self.logger, self.conf, self.names)
            self.running = dels
            dels.run()


class BenchDELETE(Bench):

    def __init__(self, logger, conf, names):
        Bench.__init__(self, logger, conf, names)
        self.concurrency = self.del_concurrency
        self.total = len(names)
        self.msg = 'DEL'

    def _run(self, thread):
        if time.time() - self.heartbeat >= 15:
            self.heartbeat = time.time()
            self._log_status('DEL')
        device, partition, name, container_name = self.names.pop()
        with self.connection() as conn:
            try:
                if self.use_proxy:
                    client.delete_object(self.url, self.token,
                                         container_name, name, http_conn=conn)
                else:
                    node = {'ip': self.ip, 'port': self.port, 'device': device}
                    direct_client.direct_delete_object(node, partition,
                                                       self.account,
                                                       container_name, name)
            except (client.ClientException,
                    requests.exceptions.ConnectionError) as e:
                self.logger.debug(str(e))
                self.failures += 1
        self.complete += 1


class BenchGET(Bench):

    def __init__(self, logger, conf, names):
        Bench.__init__(self, logger, conf, names)
        self.concurrency = self.get_concurrency
        self.total = self.total_gets
        self.msg = 'GETS'

    def _run(self, thread):
        if time.time() - self.heartbeat >= 15:
            self.heartbeat = time.time()
            self._log_status('GETS')
        device, partition, name, container_name = random.choice(self.names)
        with self.connection() as conn:
            try:
                if self.use_proxy:
                    headers, body = client.get_object(
                        self.url, self.token,
                        container_name, name, http_conn=conn,
                        resp_chunk_size=2**20)
                    with closing(body):
                        for _ in body:
                            pass
                else:
                    node = {'ip': self.ip, 'port': self.port, 'device': device}
                    direct_client.direct_get_object(node, partition,
                                                    self.account,
                                                    container_name, name)
            except (client.ClientException,
                    requests.exceptions.ConnectionError) as e:
                self.logger.debug(str(e))
                self.failures += 1
        self.complete += 1


class BenchPUT(Bench):

    def __init__(self, logger, conf, names):
        Bench.__init__(self, logger, conf, names)
        self.concurrency = self.put_concurrency
        self.total = self.total_objects
        self.msg = 'PUTS'
        self.containers = conf.containers

    def _run(self, thread):
        if time.time() - self.heartbeat >= 15:
            self.heartbeat = time.time()
            self._log_status('PUTS')
        name = uuid.uuid4().hex
        if self.object_sources:
            source = random.choice(self.files)
        elif self.upper_object_size > self.lower_object_size:
            source = SourceFile(random.randint(self.lower_object_size,
                                               self.upper_object_size))
        else:
            source = SourceFile(self.object_size)
        device = random.choice(self.devices)
        partition = str(random.randint(1, 3000))
        container_name = random.choice(self.containers)
        with self.connection() as conn:
            try:
                if self.use_proxy:
                    client.put_object(self.url, self.token,
                                      container_name, name, source,
                                      content_length=len(source),
                                      http_conn=conn)
                else:
                    node = {'ip': self.ip, 'port': self.port, 'device': device}
                    direct_client.direct_put_object(node, partition,
                                                    self.account,
                                                    container_name, name,
                                                    source,
                                                    content_length=len(source))
            except (client.ClientException,
                    requests.exceptions.ConnectionError) as e:
                self.logger.debug(str(e))
                self.failures += 1
            else:
                self.names.append((device, partition, name, container_name))
        self.complete += 1



# Copyright (c) 2022 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from unittest import mock
from swiftbench import cli


class TestCli(unittest.TestCase):

    def setUp(self):
        self.container_options = None
        self.delete_options = None

    def fake_create_containers(self, logger, options):
        self.container_options = options

    def fake_delete_containers(self, logger, options):
        self.delete_options = options

    def run_main(self, args):
        mock_controller = mock.Mock()

        with mock.patch('swiftbench.cli.create_containers',
                        self.fake_create_containers), \
                mock.patch('swiftbench.cli.delete_containers',
                           self.fake_delete_containers), \
                mock.patch('swiftbench.cli.BenchController',
                           mock_controller), \
                mock.patch('swiftbench.cli.DistributedBenchController',
                           mock_controller):
            cli.main(args)
        return (mock_controller.call_args[0][-1], self.container_options,
                self.delete_options)

    def test_defaults(self):
        controller_opts, container_opts, del_opts = self.run_main([])
        self.assertFalse(controller_opts.saio)
        self.assertEqual(controller_opts.auth, '')
        self.assertEqual(controller_opts.user, '')
        self.assertEqual(controller_opts.key, '')
        self.assertEqual(controller_opts.url, '')
        self.assertIsNone(controller_opts.concurrency)
        self.assertEqual(controller_opts.get_concurrency, 10)
        self.assertEqual(controller_opts.put_concurrency, 10)
        self.assertEqual(controller_opts.lower_object_size, 10)
        self.assertEqual(controller_opts.upper_object_size, 10)
        self.assertEqual(controller_opts.num_objects, 1000)
        self.assertEqual(controller_opts.num_gets, 10000)
        self.assertEqual(controller_opts.num_containers, 20)
        self.assertEqual(len(controller_opts.containers), 20)
        self.assertTrue(controller_opts.delete)
        self.assertEqual(controller_opts.auth_version, '1.0')
        self.assertEqual(controller_opts.delay, 0)
        self.assertIsNone(controller_opts.policy_name)
        self.assertTrue(controller_opts.use_proxy)
        self.assertEqual(controller_opts.del_concurrency, 10)
        self.assertEqual(controller_opts.object_sources, '')
        self.assertEqual(controller_opts.account, '')
        self.assertEqual(controller_opts.container_name, mock.ANY)
        self.assertEqual(controller_opts.devices, 'sdb1')
        self.assertEqual(controller_opts.log_level, 'INFO')
        self.assertEqual(controller_opts.timeout, 10)
        self.assertEqual(controller_opts.bench_clients, [])
        self.assertTrue(controller_opts.containers)

    def test_defaults_with_saio(self):
        controller_opts, container_opts, del_opts = self.run_main(['--saio'])
        self.assertTrue(controller_opts.saio)
        self.assertEqual(controller_opts.auth,
                         'http://localhost:8080/auth/v1.0')
        self.assertEqual(controller_opts.user, 'test:tester')
        self.assertEqual(controller_opts.key, 'testing')
        self.assertEqual(controller_opts.url, '')
        self.assertIsNone(controller_opts.concurrency)
        self.assertEqual(controller_opts.get_concurrency, 10)
        self.assertEqual(controller_opts.put_concurrency, 10)
        self.assertEqual(controller_opts.lower_object_size, 10)
        self.assertEqual(controller_opts.upper_object_size, 10)
        self.assertEqual(controller_opts.num_objects, 1000)
        self.assertEqual(controller_opts.num_gets, 10000)
        self.assertEqual(controller_opts.num_containers, 20)
        self.assertEqual(len(controller_opts.containers), 20)
        self.assertTrue(controller_opts.delete)
        self.assertEqual(controller_opts.auth_version, '1.0')
        self.assertEqual(controller_opts.delay, 0)
        self.assertIsNone(controller_opts.policy_name)
        self.assertTrue(controller_opts.use_proxy)
        self.assertEqual(controller_opts.del_concurrency, 10)
        self.assertEqual(controller_opts.object_sources, '')
        self.assertEqual(controller_opts.account, '')
        self.assertEqual(controller_opts.container_name, mock.ANY)
        self.assertEqual(controller_opts.devices, 'sdb1')
        self.assertEqual(controller_opts.log_level, 'INFO')
        self.assertEqual(controller_opts.timeout, 10)
        self.assertEqual(controller_opts.bench_clients, [])
        self.assertTrue(controller_opts.containers)

    def test_cli_options(self):
        controller_opts, container_opts, del_opts = self.run_main(
            [
                '--auth', 'http://some_url/auth/v1.0',
                '-U', 'other:user',
                '--key', 'my_key',
                '--bench-clients', '1.2.3.4:1234',
                '--url', 'http://storage.url/v1/AUTH_user',
                '--get-concurrency', '9',
                '--put-concurrency', '8',
                '--delete-concurrency', '7',
                '--object-size', '999',
                '--lower-object-size', '1',
                '--num-objects', '5000',
                '--num-gets', '4000',
                '--num-containers', '4',
                '--container-name', 'foo',
                '-x',
                '--auth_version', '2.0',
                '--delay', '10',
                '--policy-name', 'gold'])
        self.assertFalse(controller_opts.saio)
        self.assertEqual(controller_opts.auth, 'http://some_url/auth/v1.0')
        self.assertEqual(controller_opts.user, 'other:user')
        self.assertEqual(controller_opts.key, 'my_key')
        self.assertEqual(controller_opts.bench_clients, ['1.2.3.4:1234'])
        self.assertEqual(controller_opts.url,
                         'http://storage.url/v1/AUTH_user')
        self.assertEqual(controller_opts.get_concurrency, 9)
        self.assertEqual(controller_opts.put_concurrency, 8)
        self.assertEqual(controller_opts.del_concurrency, 7)
        self.assertEqual(controller_opts.object_size, 999)
        self.assertEqual(controller_opts.lower_object_size, 1)
        self.assertEqual(controller_opts.num_objects, 5000)
        self.assertEqual(controller_opts.num_gets, 4000)
        self.assertEqual(controller_opts.num_containers, 4)
        self.assertEqual(len(controller_opts.containers), 4)
        self.assertEqual([
            c.startswith('foo_') for c in controller_opts.containers
        ], [True] * 4)
        self.assertFalse(controller_opts.delete)
        self.assertEqual(controller_opts.auth_version, '2.0')
        self.assertEqual(controller_opts.delay, 10)
        self.assertEqual(controller_opts.policy_name, 'gold')
        self.assertTrue(controller_opts.use_proxy)
        self.assertEqual(controller_opts.object_sources, '')
        self.assertEqual(controller_opts.account, '')
        self.assertEqual(controller_opts.container_name, mock.ANY)
        self.assertEqual(controller_opts.devices, 'sdb1')
        self.assertEqual(controller_opts.log_level, 'INFO')
        self.assertEqual(controller_opts.timeout, 10)
        self.assertTrue(controller_opts.containers)

    def test_single_container(self):
        controller_opts, container_opts, del_opts = self.run_main([
            '--container-name', 'bar', '--num-containers', '1'])
        self.assertEqual(controller_opts.num_containers, 1)
        self.assertEqual(controller_opts.containers, ['bar'])

    def test_concurrency_overrides_get_put_delete(self):
        controller_opts, container_opts, del_opts = self.run_main(
            [
                '--concurrency', '5',
                '--get-concurrency', '9',
                '--put-concurrency', '8',
                '--delete-concurrency', '7'])
        self.assertEqual(controller_opts.concurrency, 5)
        self.assertEqual(controller_opts.get_concurrency, 5)
        self.assertEqual(controller_opts.put_concurrency, 5)
        self.assertEqual(controller_opts.del_concurrency, 5)



# Copyright (c) 2022 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import copy
import logging
import os
import sys
import signal
import uuid

from swiftbench.bench import (BenchController, DistributedBenchController,
                              create_containers, delete_containers)
from swiftbench.utils import readconf, config_true_value, get_size_bytes

# The defaults should be sufficient to run swift-bench on a SAIO
CONF_DEFAULTS = {
    'auth': os.environ.get('ST_AUTH', ''),
    'user': os.environ.get('ST_USER', ''),
    'key': os.environ.get('ST_KEY', ''),
    'auth_version': '1.0',
    'use_proxy': 'yes',
    'put_concurrency': 10,
    'get_concurrency': 10,
    'del_concurrency': 10,
    'object_sources': '',  # set of file contents to read and use for PUTs
    'lower_object_size': 10,  # bounded random size used if these differ
    'upper_object_size': 10,
    'object_size': 1,  # only if not object_sources and lower == upper
    'num_objects': 1000,
    'num_gets': 10000,
    'delete': 'yes',
    'container_name': uuid.uuid4().hex,  # really "container name base"
    'num_containers': 20,
    'url': '',  # used when use_proxy = no or overrides auth X-Storage-Url
    'account': '',  # used when use_proxy = no
    'devices': 'sdb1',  # space-sep list
    'log_level': 'INFO',
    'timeout': 10,
    'delay': 0,
    'bench_clients': [],
}

SAIO_DEFAULTS = {
    'auth': 'http://localhost:8080/auth/v1.0',
    'user': 'test:tester',
    'key': 'testing',
}


def main(argv):
    usage = "usage: %(prog)s [OPTIONS] [CONF_FILE]"
    usage += """\n\nConf file with SAIO defaults:

    [bench]
    auth = http://localhost:8080/auth/v1.0
    user = test:tester
    key = testing
    concurrency = 10
    object_size = 1
    num_objects = 1000
    num_gets = 10000
    delete = yes
    auth_version = 1.0
    policy_name = gold
    """
    parser = argparse.ArgumentParser(usage=usage)
    parser.add_argument('--saio', action='store_true',
                        help='Run benchmark with SAIO defaults')
    parser.add_argument('-A', '--auth',
                        help='URL for obtaining an auth token')
    parser.add_argument('-U', '--user',
                        help='User name for obtaining an auth token')
    parser.add_argument('-K', '--key',
                        help='Key for obtaining an auth token')
    parser.add_argument('-b', '--bench-clients', action='append',
                        metavar='<ip>:<port>',
                        help=('A string of the form "<ip>:<port>" which '
                              'matches the arguments supplied to a '
                              'swift-bench-client process.  This argument '
                              'must be specified once per swift-bench-client '
                              'you want to utilize.'))
    parser.add_argument('-u', '--url',
                        help='Storage URL')
    parser.add_argument('-c', '--concurrency', type=int,
                        help=('Number of concurrent connections to use. For '
                              'finer-grained control, see --get-concurrency, '
                              '--put-concurrency, and --delete-concurrency.'))
    parser.add_argument('--get-concurrency', type=int,
                        help='Number of concurrent GET requests')
    parser.add_argument('--put-concurrency', type=int,
                        help='Number of concurrent PUT requests')
    parser.add_argument('--delete-concurrency', type=int,
                        dest="del_concurrency",
                        help='Number of concurrent DELETE requests')
    parser.add_argument('-s', '--object-size', type=get_size_bytes,
                        help='Size of objects to PUT (in bytes)')
    parser.add_argument('-l', '--lower-object-size', type=get_size_bytes,
                        help=('Lower size of objects (in bytes); '
                              '--object-size will be upper-object-size'))
    parser.add_argument('-n', '--num-objects', type=int,
                        help='Number of objects to PUT')
    parser.add_argument('-g', '--num-gets', type=int,
                        help='Number of GET operations to perform')
    parser.add_argument('-C', '--num-containers', type=int,
                        help='Number of containers to distribute objects '
                             'among')
    parser.add_argument('--container-name',
                        help='Set the base container_name')
    parser.add_argument('-x', '--no-delete',
                        dest='delete', action='store_false',
                        help='If set, will not delete the objects created')
    parser.add_argument('-V', '--auth_version',
                        help='Authentication version')
    parser.add_argument('-d', '--delay', type=int,
                        help='Delay before delete requests in seconds')
    parser.add_argument('-P', '--policy-name',
                        help='Specify which policy to use when creating '
                             'containers')
    parser.add_argument('conf_file', nargs="?",
                        help='config file')

    args = parser.parse_args(argv)
    parser_defaults = copy.deepcopy(CONF_DEFAULTS)
    if args.saio:
        parser_defaults.update(SAIO_DEFAULTS)
    if getattr(args, 'lower_object_size', None):
        if args.object_size <= args.lower_object_size:
            raise ValueError('--lower-object-size (%s) must be '
                             '< --object-size (%s)' %
                             (args.lower_object_size, args.object_size))
        parser_defaults['upper_object_size'] = args.object_size
    if args.conf_file:
        if not os.path.exists(args.conf_file):
            sys.exit("No such conf file: %s" % args.conf_file)
        conf = readconf(args.conf_file, 'bench', log_name='swift-bench',
                        defaults=parser_defaults)
        conf['bench_clients'] = []
    else:
        conf = parser_defaults
    parser.set_defaults(**conf)
    options = parser.parse_args(argv)

    if options.concurrency:
        options.put_concurrency = options.concurrency
        options.get_concurrency = options.concurrency
        options.del_concurrency = options.concurrency
    if options.num_containers == 1:
        options.containers = [options.container_name]
    else:
        options.containers = ['%s_%d' % (options.container_name, i)
                              for i in range(int(options.num_containers))]

    # Turn "yes"/"no"/etc. strings to booleans
    options.use_proxy = config_true_value(options.use_proxy)
    options.delete = config_true_value(options.delete)

    def sigterm(signum, frame):
        sys.exit('Termination signal received.')
    signal.signal(signal.SIGTERM, sigterm)

    logger = logging.getLogger('swift-bench')
    logger.propagate = False
    logger.setLevel({
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'critical': logging.CRITICAL}.get(
            options.log_level.lower(), logging.INFO))
    loghandler = logging.StreamHandler()
    logger.addHandler(loghandler)
    logformat = logging.Formatter(
        'swift-bench %(asctime)s %(levelname)s %(message)s')
    loghandler.setFormatter(logformat)

    if options.use_proxy:
        create_containers(logger, options)

    controller_class = DistributedBenchController if options.bench_clients \
        else BenchController
    controller = controller_class(logger, options)
    controller.run()

    if options.use_proxy and options.delete:
        delete_containers(logger, options)



# Copyright (c) 2010-2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import configparser
from urllib.parse import urlparse
from urllib.request import getproxies, proxy_bypass

# Used when reading config values
TRUE_VALUES = set(('true', '1', 'yes', 'on', 't', 'y'))


# NOTE(chmouel): Imported from swift without the modular directory feature.
def readconf(conf_path, section_name=None, log_name=None, defaults=None,
             raw=False):
    """
    Read config file(s) and return config items as a dict

    :param conf_path: path to config file, or a file-like object
                     (hasattr readline)
    :param section_name: config section to read (will return all sections if
                     not defined)
    :param log_name: name to be used with logging (will use section_name if
                     not defined)
    :param defaults: dict of default values to pre-populate the config with
    :returns: dict of config items
    """
    if defaults is None:
        defaults = {}
    if raw:
        c = configparser.RawConfigParser(defaults)
    else:
        c = configparser.ConfigParser(defaults)
    if hasattr(conf_path, 'readline'):
        # make sure filelike is at the start of the file
        if hasattr(conf_path, "seekable"):
            conf_path.seek(0)
        c.read_file(conf_path)
    else:
        success = c.read(conf_path)
        if not success:
            print("Unable to read config from %s" % conf_path)
            sys.exit(1)
    if section_name:
        if c.has_section(section_name):
            conf = dict(c.items(section_name))
        else:
            print("Unable to find %s config section in %s" %
                  (section_name, conf_path))
            sys.exit(1)
        if "log_name" not in conf:
            if log_name is not None:
                conf['log_name'] = log_name
            else:
                conf['log_name'] = section_name
    else:
        conf = {}
        for s in c.sections():
            conf.update({s: dict(c.items(s))})
        if 'log_name' not in conf:
            conf['log_name'] = log_name
    conf['__file__'] = conf_path
    return conf


def config_true_value(value):
    """
    Returns True if the value is either True or a string in TRUE_VALUES.
    Returns False otherwise.
    """
    return value is True or \
        (isinstance(value, str) and value.lower() in TRUE_VALUES)


def using_http_proxy(url):
    """
    Return True if the url will use HTTP proxy.
    Returns False otherwise.
    """
    up = urlparse(url)
    return up.scheme.lower() in getproxies() and not proxy_bypass(up.netloc)


def get_size_bytes(value):
    """
    Expand suffixes for object sizes.

    For example, '10k' becomes 10240, and '2M' becomes 2097152.
    """
    if isinstance(value, int):
        return value
    if not isinstance(value, str):
        raise TypeError
    value = value.strip()
    multiple = {
        'G': 1024 ** 3,
        'M': 1024 ** 2,
        'k': 1024,
    }.get(value[-1])
    if multiple is None:
        return int(value)
    return multiple * int(value[:-1])



#!/usr/bin/python
# Copyright (c) 2010-2012 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

from swiftbench import __version__ as version


name = 'swift-bench'

with open('requirements.txt', 'r') as f:
    requires = [x.strip() for x in f if x.strip()]


setup(
    name=name,
    version=version,
    description='Benchmark tool for OpenStack Swift',
    license='Apache License (2.0)',
    author='OpenStack',
    author_email='openstack-discuss@lists.openstack.org',
    url='http://openstack.org',
    packages=find_packages(exclude=['test', 'bin']),
    test_suite='nose.collector',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Environment :: No Input/Output (Daemon)',
        'Environment :: OpenStack',
    ],
    install_requires=requires,
    scripts=[
        'bin/swift-bench',
        'bin/swift-bench-client',
    ],
)
